/* 
 * 
 * (c) Copyright Ascensio System Limited 2010-2014
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * http://www.gnu.org/licenses/agpl.html 
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using ASC.Api.Attributes;
using ASC.Api.Projects.Wrappers;
using ASC.Api.Utils;
using ASC.Projects.Core.Domain;
using ASC.Projects.Core.Domain.Reports;
using ASC.Projects.Engine;
using ASC.Specific;

namespace ASC.Api.Projects
{
    public partial class  ProjectApi
    {
		///<summary>
		/// Creates the project report template with the parameters specified in the request. Most of the parameters are optional and depend on the report type selected.
		///</summary>
		///<short>
		/// Create report template
		///</short>
		/// <category>Report</category>         
		///<param name="name">Report name</param>
		///<param name="period" optional="true">Report template generation period. Can be one of the following: Day/Week/Month</param>
		///<param name="periodItem" optional="true">Period item. Can be: weekday name (for weekly report, starting with Sunday), day of the month (for monthly report).</param>
		///<param name="hour" optional="true">Send time (hour)</param>
		///<param name="autoGenerated" optional="true">Automatically generated or not</param>
		///<param name="reportType" optional="true">Report type</param>
		///<param name="tag" optional="true">Tag</param>
		///<param name="project" optional="true">Project</param>
		///<param name="status" optional="true">Task status</param>       
		///<param name="departament" optional="true">Department/Group</param>
		///<param name="userId" optional="true">User GUID</param>
		///<param name="reportTimeInterval" optional="true">Report time interval</param>
		///<param name="fromDate" optional="true">Report period start date</param>
		///<param name="toDate" optional="true">Report period end date</param>
		///<param name="viewType" optional="true">View type: view by group or view by project</param>
		///<param name="noResponsible" optional="true">Show or hide tasks without responsible</param>   
		///<returns>Project report template</returns>
		[Create(@"report")]
        public ReportTemplateWrapper SaveReportTemplate(string name, string period, int periodItem, int hour, bool autoGenerated, ReportType reportType,
                                         int tag, int project, TaskStatus? status, Guid departament, Guid userId, ReportTimeInterval reportTimeInterval,
                                         ApiDateTime fromDate, ApiDateTime toDate, int viewType, bool noResponsible)
        {
            ProjectSecurity.DemandAuthentication();

            if (name == null || name.Trim().Length == 0) throw new ArgumentNullException("name");

            var filter = new TaskFilter
            {
                TagId = tag,
                DepartmentId = departament,
                UserId = userId,
                TimeInterval = reportTimeInterval,
                FromDate = fromDate,
                ToDate = toDate,
                ViewType = viewType,
                NoResponsible = noResponsible
            };

            if (project != 0)
                filter.ProjectIds.Add(project);

            if (status != null)
                filter.TaskStatuses.Add((TaskStatus)status);

            var template = new ReportTemplate(reportType) { Filter = filter };

            SaveOrUpdateTemplate(template, name, period, periodItem, hour, autoGenerated);

            return new ReportTemplateWrapper(template);
        }

		///<summary>
		/// Updates the selected project report template with the parameters specified in the request
		///</summary>
		///<short>
		/// Update report template
		///</short>
		/// <category>Report</category>           
		///<param name="reportid">Report template ID</param>    
		///<param name="name">Report name</param>
		///<param name="period" optional="true">Report template generation period. Can be one of the following: Day/Week/Month</param>
		///<param name="periodItem" optional="true">Period item. Can be: weekday name (for weekly report, starting with Sunday), day of the month (for monthly report).</param>
		///<param name="hour" optional="true">Send time (hour)</param>
		///<param name="autoGenerated" optional="true">Automatically generated or not</param>
		///<param name="reportType" optional="true">Report type</param>
		///<param name="tag" optional="true">Tag</param>
		///<param name="project" optional="true">Project</param>
		///<param name="status" optional="true">Task status</param>       
		///<param name="departament" optional="true">Department/Group</param>
		///<param name="userId" optional="true">User GUID</param>
		///<param name="reportTimeInterval" optional="true">Report time interval</param>
		///<param name="fromDate" optional="true">Report period start date</param>
		///<param name="toDate" optional="true">Report period end date</param>
		///<param name="viewType" optional="true">View type: view by group or view by project</param>
		///<param name="noResponsible" optional="true">Show or hide tasks without responsible</param>   
		///<returns>Project report template</returns>
		[Update(@"report/{reportid:[0-9]+}")]
        public ReportTemplateWrapper UpdateReportTemplate(int reportid, string name, string period, int periodItem, int hour, bool autoGenerated, ReportType reportType,
                                         int tag, int project, TaskStatus? status, Guid departament, Guid userId, ReportTimeInterval reportTimeInterval,
                                         ApiDateTime fromDate, ApiDateTime toDate, int viewType, bool noResponsible)
        {
            ProjectSecurity.DemandAuthentication();

            var filter = new TaskFilter
            {
                TagId = tag,
                DepartmentId = departament,
                UserId = userId,
                TimeInterval = reportTimeInterval,
                FromDate = fromDate,
                ToDate = toDate,
                ViewType = viewType,
                NoResponsible = noResponsible
            };

            if (project != 0)
                filter.ProjectIds.Add(project);

            if (status != null)
                filter.TaskStatuses.Add((TaskStatus)status);


            var template = EngineFactory.GetReportEngine().GetTemplate(reportid);
            template.Filter = filter;

            SaveOrUpdateTemplate(template, name, period, periodItem, hour, autoGenerated);
            return new ReportTemplateWrapper(template);
        }

		///<summary>
		/// Deletes the project report template with the ID specified in the request
		///</summary>
		///<short>
		/// Delete report template
		///</short>
		/// <category>Report</category>   
		///<param name="reportid">Report template ID</param>        
		///<returns>Project report template</returns>
		[Delete(@"report/{reportid:[0-9]+}")]
        public ReportTemplateWrapper DeleteReportTemplate(int reportid)
        {
            ProjectSecurity.DemandAuthentication();

            var reportEngine = EngineFactory.GetReportEngine();

            var reportTemplate = reportEngine.GetTemplate(reportid).NotFoundIfNull();

            reportEngine.DeleteTemplate(reportid);

            return new ReportTemplateWrapper(reportTemplate);
        }

        private void SaveOrUpdateTemplate(ReportTemplate template, string title, string period, int periodItem, int hour, bool autoGenerated)
        {
            template.Name = HttpUtility.HtmlEncode(title);
            switch (period)
            {
                case "day":
                    template.Cron = string.Format("0 0 {0} * * ?", hour);
                    break;
                case "week":
                    template.Cron = string.Format("0 0 {0} ? * {1}", hour, periodItem);
                    break;
                case "month":
                    template.Cron = string.Format("0 0 {0} {1} * ?", hour, periodItem);
                    break;
                default:
                    template.Cron = string.Format("0 0 {0} * * ?", 12);
                    break;
            }
            template.AutoGenerated = autoGenerated;
            EngineFactory.GetReportEngine().SaveTemplate(template);
        }
    }
}
